@@ -208,6 +208,12 @@ main(int argc, const char * const argv[])
 	socklen_t saslength = sizeof(sas);
 	long client[2] = { 0, 0 };
 	long server[2] = { 0, 0 };
+    int useIPv6=0;
+    int useIPv4=0;
+    int preferIPv4=0;
+    int preferIPv6=0;
+    int tryIPv4=0;
+    int tryIPv6=0;
 
 	rv = apr_app_initialize(&argc, &argv, NULL);
 
@@ -255,19 +261,23 @@ main(int argc, const char * const argv[])
 			{ "sslsink", no_argument, 0, 's' },
 			{ "heartbeat", no_argument, 0, 'h' },
 			{ "debug", no_argument, 0, 'd' },
+			{ "useIPv4", no_argument, 0, '4' },
+			{ "useIPv6", no_argument, 0, '6' },
+			{ "preferIPv4", no_argument, 0, 'y' },
+			{ "preferIPv6", no_argument, 0, 'z' },
 			{ 0, 0, 0, 0 }
 		};
 
 		int option_index = 0;
 
-		c = getopt_long(argc, (char **) argv, "r?vqp:b:shd", long_options, &option_index);
+		c = getopt_long(argc, (char **) argv, "r?vqp:b:shd46yz", long_options, &option_index);
 
 		if (c == -1) 
 			break;
 
 		switch (c) {
-			case 'r': fprintf(stdout, "vncproxy 2.1\n");
-					  fprintf(stdout, "Copyright (c) 2010-2012 Purdue University\n");
+			case 'r': fprintf(stdout, "vncproxy 2.2.3\n");
+					  fprintf(stdout, "Copyright (c) 2010-2013 Purdue University\n");
 					  fprintf(stdout, "This is free software: you can redistribute it and/or modify it under the terms\n");
 					  fprintf(stdout, "of the GNU Lesser General Public License version 3 or later.\n");
 					  fprintf(stdout, "<http://gnu.org/licenses/gpl.html>\n"); 
@@ -293,6 +303,14 @@ main(int argc, const char * const argv[])
 					  fprintf(stdout, "  -h, --heartbeat        maintain heartbeat field in view table\n");
 					  fprintf(stdout, "  -d, --debug            print debugging information to stderr\n");
 					  fprintf(stdout, "\n");
+					  fprintf(stdout, "If none of the following options are selected, host resolution is done\n");
+					  fprintf(stdout, "in accordance with the current operating system defaults (typically \n");
+					  fprintf(stdout, "something similar to --preferIPv6)\n");
+					  fprintf(stdout, "  --useIPv4,    -4       lookup remote hosts using only IPv4\n");
+					  fprintf(stdout, "  --useIPv6,    -6       lookup remote hosts using only IPv6\n");
+					  fprintf(stdout, "  --preferIPv4, -y       lookup remote hosts using IPv4 first\n");
+					  fprintf(stdout, "  --preferIPv6, -z       lookup remote hosts using IPv6 first\n");
+					  fprintf(stdout, "\n");
 					  fprintf(stdout, "  --help                 display this help and exit\n");
 					  fprintf(stdout, "  --version              output version information and exit\n");
 					  fprintf(stdout, "\n");
@@ -304,6 +322,10 @@ main(int argc, const char * const argv[])
 			case 'v': verbose = 1; break;
 			case 'q': quiet = 1; break;
 			case 'd': debug = 1; break;
+			case '4': useIPv4 = 1; break;
+			case '6': useIPv6 = 1; break;
+			case 'y': preferIPv4 = 1; break;
+			case 'z': preferIPv6 = 1; break;
 			default: break;
 		}
 	}
@@ -314,7 +336,7 @@ main(int argc, const char * const argv[])
 
 	if (token == NULL || *token == 0 || optind < argc) {
 		fprintf(stderr, "vncproxy: invalid option or token\n");
-		fprintf(stderr,"Try `vncproxy --help` for more information.\n");
+		fprintf(stderr, "Try `vncproxy --help` for more information.\n");
 		exit(5);
 	}
 
@@ -334,6 +356,12 @@ main(int argc, const char * const argv[])
 			log_priority = LOG_NOTICE;
 	}
 
+    if (useIPv4 + useIPv6 + preferIPv4 + preferIPv6 > 1) {
+    	fprintf(stderr, "vncproxy: conflicting options\n");
+    	fprintf(stderr, "Use only one of --useIPv4, --useIPv6, --preferIPv4, --preferIPv6\n");
+    	exit(6);
+    }
+
 	sigemptyset(&sigmask);
 	sigaddset(&sigmask, SIGHUP);
 	sigaddset(&sigmask, SIGTERM);
@@ -342,22 +370,22 @@ main(int argc, const char * const argv[])
 
 	if (signal(SIGHUP, do_proxy_vnc_sighandler) == SIG_ERR) {
 		logmsg(MLOG_ERR, "Failed to assign signal handler to SIGHUP");
-		exit(6);
+		exit(7);
 	}
 
 	if (signal(SIGTERM, do_proxy_vnc_sighandler) == SIG_ERR) {
 		logmsg(MLOG_ERR, "Failed to assign signal handler to SIGTERM");
-		exit(7);
+		exit(8);
 	}
 
 	if (signal(SIGINT, do_proxy_vnc_sighandler) == SIG_ERR) {
 		logmsg(MLOG_ERR, "Failed to assign signal handler to SIGINT");
-		exit(8);
+		exit(9);
 	}
 
 	if (signal(SIGQUIT, do_proxy_vnc_sighandler) == SIG_ERR) {
 		logmsg(MLOG_ERR, "Failed to assign signal handler to SIGINT");
-		exit(9);
+		exit(10);
 	}
 
 	logmsg(MLOG_INFO, "Starting vncproxy");
@@ -372,7 +400,7 @@ main(int argc, const char * const argv[])
 
 	if (dbd_driver == NULL) {
 		logmsg(MLOG_ERR, "Unable to allocate memory for DBD_DRIVER variable");
-		exit(10);
+		exit(11);
 	}
 
 	dbd_params = getenv("DBD_PARAMS");
@@ -385,14 +413,14 @@ main(int argc, const char * const argv[])
 
 	if (dbd_params == NULL) {
 		logmsg(MLOG_ERR, "Unable to allocate memory for DBD_PARAMS variable");
-		exit(11);
+		exit(12);
 	}
 
 	rv = apr_dbd_get_driver(pool, dbd_driver, &driver);
 
 	if (rv != APR_SUCCESS) {
 		logmsg(MLOG_ERR, "Unable to load APR_DBD driver [%s]", dbd_driver);
-		exit(12);
+		exit(13);
 	}
 
 	// free(dbd_driver);
@@ -401,7 +429,7 @@ main(int argc, const char * const argv[])
 
 	if (rv != 0) {
 		logmsg(MLOG_ERR, "Unable to open database to lookup token");
-		exit(13);
+		exit(14);
 	}
 
 	etoken = apr_dbd_escape(driver, pool, token, handle);
@@ -409,7 +437,7 @@ main(int argc, const char * const argv[])
 	if (etoken == NULL) {
 		logmsg(MLOG_ERR, "Unable to build query safe token string");
 		apr_dbd_close(driver, handle);
-		exit(14);
+		exit(15);
 	}
 
 	getpeername(STDIN_FILENO, (struct sockaddr *) &sas, &saslength);
@@ -434,7 +462,7 @@ main(int argc, const char * const argv[])
 	if (eremoteip == NULL) {
 		logmsg(MLOG_ERR, "Unable to build query safe remoteip string");
 		apr_dbd_close(driver, handle);
-		exit(15);
+		exit(16);
 	}
 
 	rv = snprintf(query, sizeof (query), "SELECT viewperm.sessnum,viewuser,display.hostname,session.dispnum,"
@@ -448,7 +476,7 @@ main(int argc, const char * const argv[])
 	if (rv < 0 || rv >= sizeof (query)) {
 		logmsg(MLOG_ERR, "Failed to build query string");
 		apr_dbd_close(driver, handle);
-		exit(16);
+		exit(17);
 	}
 
 	rv = apr_dbd_select(driver, pool, handle, &res, query, 0);
@@ -460,7 +488,7 @@ main(int argc, const char * const argv[])
 	if ((rv != APR_SUCCESS) || (apr_dbd_get_row(driver, pool, res, &row, -1) != 0)) {
 		logmsg(MLOG_WARNING, "No database entry found for token [%s]", token);
 		apr_dbd_close(driver, handle);
-		exit(17);
+		exit(18);
 	}
 
 	value = apr_dbd_get_entry(driver, row, 0);
@@ -468,7 +496,7 @@ main(int argc, const char * const argv[])
 	if (value == NULL) {
 		logmsg(MLOG_ERR, "Unable to read sessnum column from result row");
 		apr_dbd_close(driver, handle);
-		exit(18);
+		exit(19);
 	}
 
 	sessnum = atoi(value);
@@ -476,7 +504,7 @@ main(int argc, const char * const argv[])
 	if (sessnum <= 0) {
 		logmsg(MLOG_ERR, "Invalid sessnum column read from result row");
 		apr_dbd_close(driver, handle);
-		exit(19);
+		exit(20);
 	}
 
 	value = apr_dbd_get_entry(driver, row, 1);
@@ -484,7 +512,7 @@ main(int argc, const char * const argv[])
 	if (value == NULL) {
 		logmsg(MLOG_ERR, "Unable to read viewuser column from result row");
 		apr_dbd_close(driver, handle);
-		exit(20);
+		exit(21);
 	}
 
 	strncpy(username, value, sizeof (username));
@@ -494,7 +522,7 @@ main(int argc, const char * const argv[])
 	if (value == NULL) {
 		logmsg(MLOG_ERR, "Unable to read hostname column from result row");
 		apr_dbd_close(driver, handle);
-		exit(21);
+		exit(22);
 	}
 
 	strncpy(hostname, value, sizeof (hostname));
@@ -504,7 +532,7 @@ main(int argc, const char * const argv[])
 	if (value == NULL) {
 		logmsg(MLOG_ERR, "Unable to read dispnum column from result row");
 		apr_dbd_close(driver, handle);
-		exit(22);
+		exit(23);
 	}
 
 	dispnum = atoi(value);
@@ -512,7 +540,7 @@ main(int argc, const char * const argv[])
 	if (dispnum <= 0) {
 		logmsg(MLOG_ERR, "Invalid dispnum column read from result row");
 		apr_dbd_close(driver, handle);
-		exit(23);
+		exit(24);
 	}
 
 	value = apr_dbd_get_entry(driver, row, 4);
@@ -520,7 +548,7 @@ main(int argc, const char * const argv[])
 	if (value == NULL) {
 		logmsg(MLOG_ERR, "Unable to read timeout column from result row");
 		apr_dbd_close(driver, handle);
-		exit(24);
+		exit(25);
 	}
 
 	timeout = atoi(value);
@@ -528,7 +556,7 @@ main(int argc, const char * const argv[])
 	if (timeout <= 0) {
 		logmsg(MLOG_ERR, "Invalid timeout column read from result row");
 		apr_dbd_close(driver, handle);
-		exit(25);
+		exit(26);
 	}
 
 	port = dispnum + portbase;
@@ -540,7 +568,7 @@ main(int argc, const char * const argv[])
 	if (rv != -1) {
 		logmsg(MLOG_ERR, "Database transaction not finished when expected.");
 		apr_dbd_close(driver, handle);
-		exit(26);
+		exit(27);
 	}
 
 	rv = gettimeofday(&tv, NULL);
@@ -556,7 +584,7 @@ main(int argc, const char * const argv[])
 	if (rv < 0 || rv >= sizeof (query)) {
 		logmsg(MLOG_ERR, "Failed to build query string");
 		apr_dbd_close(driver, handle);
-		exit(27);
+		exit(28);
 	}
 
 	rv = apr_dbd_query(driver, handle, &nrows, query);
@@ -566,7 +594,7 @@ main(int argc, const char * const argv[])
 	if (rv != APR_SUCCESS) {
 		logmsg(MLOG_ERR, "Failed to update session accesstime");
 		apr_dbd_close(driver, handle);
-		exit(28);
+		exit(29);
 	}
 
 	memset(&hints, 0, sizeof(hints));
@@ -579,37 +607,78 @@ main(int argc, const char * const argv[])
 
 	if (rv != 0) {
 		logmsg(MLOG_ERR, "Unable to get host lookup information for %s", hostname);
-		exit(29);
+		exit(30);
 	}
 
-	for (tmpaddr = hostaddr; tmpaddr; tmpaddr = tmpaddr->ai_next) {
-		if (tmpaddr->ai_family == AF_UNIX)
-			continue;
+    tmpaddr = hostaddr;
 
-    	if ((fd = socket(tmpaddr->ai_family, SOCK_STREAM, 0)) < 0) {
-			logmsg(MLOG_ERR,"unable to create socket [%d]", errno);
-			freeaddrinfo(hostaddr);
-        	exit(30);
-    	}
+    if (useIPv6 || preferIPv6) {
+    	tryIPv4 = 0;
+    	tryIPv6 = 1;
+    } 
+    else if (useIPv4 || preferIPv4) {
+    	tryIPv4 = 1;
+    	tryIPv6 = 0;
+    }
+    else {
+    	tryIPv4 = 1;
+    	tryIPv6 = 1;
+    }
+
+	while(tmpaddr) {
+
+		if (((tmpaddr->ai_family == AF_INET6) && tryIPv6) || ((tmpaddr->ai_family == AF_INET) && tryIPv4))
+		{
+    		if ((fd = socket(tmpaddr->ai_family, SOCK_STREAM, 0)) < 0) {
+				logmsg(MLOG_ERR,"socket(AF=%d, SOCK_STREAM:=%d, 0) %s [%d]", tmpaddr->ai_family,SOCK_STREAM,strerror(errno), errno);
+    		}
+			else {		
+				((struct sockaddr_in *) tmpaddr->ai_addr)->sin_port = htons(port);
+
+				do {
+					errno = 0;
+					rv = connect(fd, tmpaddr->ai_addr, tmpaddr->ai_addrlen);
+				}
+				while (rv != 0 && errno == EINTR);
+
+				logmsg(MLOG_NOTICE,"connect(%d, AF=%d, %s:%d, %d): %s [%d]", fd, tmpaddr->ai_family, hostname, port, tmpaddr->ai_addrlen, strerror(errno), errno);
+
+	 			if (rv == 0)
+ 					break;
 		
-		((struct sockaddr_in *) tmpaddr->ai_addr)->sin_port = htons(port);
+		        shutdown(fd, SHUT_RDWR);
+        		close(fd);
+    		}
+   		}
 
-		do {
-			rv = connect(fd, tmpaddr->ai_addr, tmpaddr->ai_addrlen);
-		}
-		while (rv != 0 && errno == EINTR);
+    	tmpaddr = tmpaddr->ai_next;
 
-		if (rv != 0) {
-			logmsg(MLOG_ERR,"connect(%d, AF=%d, %s:%d, %d): %s [%d]", fd, tmpaddr->ai_family, hostname, port, tmpaddr->ai_addrlen, strerror(errno), errno);
-        	shutdown(fd, SHUT_RDWR);
-        	close(fd);
-			freeaddrinfo(hostaddr);
-			exit(31);
+    	if (tmpaddr == NULL) {
+    		if (preferIPv4 && tryIPv4) {
+    			tryIPv4 = 0;
+    			tryIPv6 = 1;
+    			tmpaddr = hostaddr;
+    		}
+    		else if (preferIPv6 && tryIPv6) {
+    			tryIPv4 = 1;
+    			tryIPv6 = 1;
+    			tmpaddr = hostaddr;
+    		}
+    		else
+    		{
+    			tryIPv4 = 0;
+    			tryIPv6 = 0;
+    		}
     	}
 	}
 
 	freeaddrinfo(hostaddr);
 
+	if (rv != 0) {
+		logmsg(MLOG_ERR,"Failed to connect to host %s.", hostname);
+		exit(31);
+	}
+
 	dup2(fd,1);
 	close(fd);
 
